""" Configuration module """
__all__ = ['config', 'logger', 'ROOT_FOLDER']

from datoso import ROOT_FOLDER

from datoso.configuration.configuration import config
from datoso.configuration.logger import logger
